#
# See
# https://discourse.julialang.org/t/is-it-possible-to-detect-if-julia-is-ahead-of-time-precompiling/78631
#
is_precompiling() = ccall(:jl_generating_output, Cint, ()) == 1




"""
$(SIGNATURES)

Solve time step problem. This is the core routine
for implicit Euler and stationary solve
"""
function _solve_timestep!(
    solution::AbstractMatrix{Tv}, # old time step solution resp. initial value
    oldsol::AbstractMatrix{Tv}, # old time step solution resp. initial value
    system::AbstractSystem{Tv,Tc,Ti,Tm}, # Finite volume system
    control::SolverControl,
    time,
    tstep,
    embedparam,
    params;
    called_from_API = false,
) where {Tv,Tc,Ti,Tm}
    _complete!(system; create_newtonvectors = true)
    nlhistory = NewtonSolverHistory()
    tasm = 0.0
    tlinsolve = 0.0
    t = @elapsed begin
        solution .= oldsol
        residual = system.residual
        update = system.update
        _initialize!(solution, system; time, λ = embedparam, params)

        method_linear = control.method_linear
        if isnothing(method_linear)
            if Tv != Float64
                method_linear = SparspakFactorization()
            else
                if dim_grid(system.grid) == 1
                    method_linear = KLUFactorization()
                elseif dim_grid(system.grid) == 2
                    method_linear = SparspakFactorization()
                else
                    method_linear = UMFPACKFactorization()
                end
            end
        end

        oldnorm = 1.0
        converged = false
        damp = 1.0
        if !system.is_linear
            if doprint(control, 'n')
                println("\n  [n]ewton: #it(lin)  |update| cont3tion   |round| #rd")
            end
            damp = control.damp_initial
            rnorm = control.rnorm(solution)
        end

        nlu_reuse = 0
        nround = 0
        tolx = 0.0
        ncalloc = 0
        nballoc = 0
        neval = 0
        niter = 1

        while niter <= control.maxiters
            # Create Jacobi matrix and RHS for Newton iteration
            try
                tasm += @elapsed nca, nba, nev = eval_and_assemble(
                    system,
                    solution,
                    oldsol,
                    residual,
                    time,
                    tstep,
                    embedparam,
                    params;
                    edge_cutoff = control.edge_cutoff,
                )
                ncalloc += nca
                nballoc += nba
                neval += nev
            catch err
                if (control.handle_exceptions)
                    _print_error(err, stacktrace(catch_backtrace()))
                    throw(AssemblyError())
                else
                    rethrow(err)
                end
            end

            tlinsolve += @elapsed _solve_linear!(
                values(update),
                system,
                nlhistory,
                control,
                method_linear,
                system.matrix,
                values(residual),
            )

            values(solution) .-= damp * values(update)
            
            # "incremental collection may only sweep   so-called young objects"
            GC.gc(false)

            if system.is_linear
                converged = true
                break
            end

            damp = min(damp * control.damp_growth, 1.0)
            norm = control.unorm(update)
            if tolx == 0.0
                tolx = norm * control.reltol
            end
            dnorm = 1.0
            rnorm_new = control.rnorm(solution)
            if rnorm > 1.0e-50
                dnorm = abs((rnorm - rnorm_new) / rnorm)
            end

            if dnorm < control.tol_round
                nround = nround + 1
            else
                nround = 0
            end

            if control.log
                push!(nlhistory.l1normdiff, dnorm)
                push!(nlhistory.updatenorm, norm)
            end
            if doprint(control, 'n')
                if control.reltol_linear < 1.0
                    itstring = @sprintf("  [n]ewton: % 3d(% 3d)", niter, nlhistory.nlin)
                else
                    itstring = @sprintf("it=% 3d", niter)
                end
                if control.max_round > 0
                    @printf(
                        "%s %.3e %.3e %.3e % 2d\n",
                        itstring,
                        norm,
                        norm / oldnorm,
                        dnorm,
                        nround
                    )
                else
                    @printf("%s %.3e %.3e\n", itstring, norm, norm / oldnorm)
                end
            end
            if niter > 1 && norm / oldnorm > 1.0 / control.tol_mono
                converged = false
                break
            end

            if norm < control.abstol || norm < tolx
                converged = true
                break
            end
            oldnorm = norm
            rnorm = rnorm_new

            if nround > control.max_round
                converged = true
                break
            end
            niter = niter + 1
        end
        if !converged
            throw(ConvergenceError())
        end
    end
    if control.log
        nlhistory.time = t
        nlhistory.tlinsolve = tlinsolve
        nlhistory.tasm = tasm
    end

    if ncalloc + nballoc > 0 && doprint(control, 'a') && !is_precompiling()
        @warn "[a]llocations in assembly loop: cells: $(ncalloc÷neval), bfaces: $(nballoc÷neval)"
    end

    if doprint(control, 'n') && !system.is_linear
        println(
            "  [n]ewton: $(round(t,sigdigits=3)) seconds asm: $(round(100*tasm/t,sigdigits=3))%, linsolve: $(round(100*tlinsolve/t,sigdigits=3))%",
        )
    end

    if doprint(control, 'l') && system.is_linear
        println(
            "  [l]inear($(nameof(typeof(method_linear)))): $(round(t,sigdigits=3)) seconds",
        )
    end

    system.history = nlhistory
end

################################################################
"""
````
solve!(solution, inival, system; 
    control=SolverControl(), 
    tstep=Inf)
````
Mutating version of [`solve(inival,system)`](@ref)
"""
function VoronoiFVM.solve!(
    solution, # Solution
    inival,   # Initial value 
    system::VoronoiFVM.AbstractSystem;     # Finite volume system
    control = SolverControl(),      # Newton solver control information
    time = Inf,
    tstep = Inf,                # Time step size. Inf means  stationary solution
    embedparam = 0.0,
    params = zeros(0),
    called_from_API = false,
)
    fix_deprecations!(control)
    if !called_from_API && doprint(control, 'd')
        @warn "[d]eprecated: solve(inival,solution, system; kwargs...)"
    end
    _solve_timestep!(
        solution,
        inival,
        system,
        control,
        time,
        tstep,
        embedparam,
        params;
        called_from_API = true,
    )
    return solution
end

################################################################
"""
````
    solve(inival, system; control=SolverControl(),params, tstep=Inf)
````
Alias for [`solve(system::VoronoiFVM.AbstractSystem; kwargs...)`](@ref) with the corresponding keyword arguments.

Solve stationary problem(if `tstep==Inf`) or one step implicit Euler step using Newton's method with `inival` as initial
value. Returns a solution array.
"""
function CommonSolve.solve(
    inival,   # Initial value 
    system::AbstractSystem;     # Finite volume system
    control = SolverControl(),      # Newton solver control information
    time = Inf,
    tstep = Inf,                # Time step size. Inf means  stationary solution
    params = zeros(0),
    called_from_API = false,
)
    fix_deprecations!(control)
    if !called_from_API && doprint(control, 'd')
        @warn "[d]eprecated: solve(inival,system; kwargs...)"
    end

    solve!(
        unknowns(system),
        inival,
        system;
        control = control,
        time = time,
        tstep = tstep,
        params = params,
        called_from_API = true,
    )
end

Δλ_val(control, transient) = transient ? control.Δt : control.Δp
Δλ_min(control, transient) = transient ? control.Δt_min : control.Δp_min
Δλ_max(control, transient) = transient ? control.Δt_max : control.Δp_max
Δλ_grow(control, transient) = transient ? control.Δt_grow : control.Δp_grow

"""
        solve(inival, system, times; kwargs...)

Alias for [`solve(system::VoronoiFVM.AbstractSystem; kwargs...)`](@ref) with the corresponding keyword arguments.

"""
function CommonSolve.solve(
    inival,
    system::VoronoiFVM.AbstractSystem,
    lambdas;
    control = SolverControl(),
    transient = true, # choose between transient and stationary (embedding) case
    time = 0.0,
    params = zeros(0),
    called_from_API = false,
    kwargs...,
)
    fix_deprecations!(control)
    if !called_from_API && doprint(control, 'd')
        @warn "[d]eprecated: solve(inival,system,times;kwargs...)"
    end
    λstr = "t"
    if !transient
        λstr = "p"
    end

    allhistory = TransientSolverHistory()

    solution = copy(inival)
    oldsolution = copy(inival)
    _initialize_dirichlet!(inival, system; time, λ = Float64(lambdas[1]), params)
    Δλ = Δλ_val(control, transient)

    t0 = @elapsed if !transient
        control.pre(solution, Float64(lambdas[1]))
        solution = solve!(
            solution,
            oldsolution,
            system;
            called_from_API = true,
            control = control,
            time = time,
            tstep = Inf,
            embedparam = Float64(lambdas[1]),
            params = params,
            kwargs...,
        )
        control.post(solution, oldsolution, lambdas[1], 0)
        if control.log
            push!(allhistory, system.history)
            push!(allhistory.times, lambdas[1])
            Δu = control.delta(system, solution, oldsolution, lambdas[1], 0)
            push!(allhistory.updates, Δu)
        end
        oldsolution .= solution
    end

    tsol = TransientSolution(Float64(lambdas[1]), solution; in_memory = control.in_memory)

    if doprint(control, 'e')
        println("[e]volution: start in $(extrema(lambdas))")
    end

    λ0 = 0
    istep = 0
    solved = false
    t1 = @elapsed for i = 1:(length(lambdas)-1)
        Δλ = max(Δλ, Δλ_min(control, transient))
        λstart = lambdas[i]
        λend = lambdas[i+1]
        λ = Float64(λstart)

        while λ < λend
            solved = false
            λ0 = λ
            Δu = 0.0
            while !solved
                solved = true
                try
                    λ = λ0 + Δλ
                    control.pre(solution, λ)
                    if transient
                        solution = solve!(
                            solution,
                            oldsolution,
                            system;
                            called_from_API = true,
                            control = control,
                            time = λ,
                            tstep = Δλ,
                            params = params,
                            kwargs...,
                        )
                    else
                        solution = solve!(
                            solution,
                            oldsolution,
                            system;
                            called_from_API = true,
                            control = control,
                            time = time,
                            tstep = Inf,
                            embedparam = λ,
                            params = params,
                            kwargs...,
                        )
                    end
                catch err
                    if (control.handle_exceptions)
                        _print_error(err, stacktrace(catch_backtrace()))
                    else
                        rethrow(err)
                    end
                    solved = false
                end
                if solved
                    Δu = control.delta(system, solution, oldsolution, λ, Δλ)
                    if Δu > 2.0 * control.Δu_opt
                        solved = false
                    end
                end
                if !solved
                    # reduce time step and retry  solution
                    Δλ = Δλ * 0.5
                    rd(x)=round(x,sigdigits=5)
                    if Δλ < Δλ_min(control, transient)
                        if !(control.force_first_step && istep == 0)
                            err="At $(λstr)=$(λ|>rd): Δ$(λstr)_min=$(Δλ_min(control,transient)|>rd) reached while Δu=$(Δu|>rd) and Δu_opt=$(control.Δu_opt|>rd) "
                            if control.handle_exceptions
                                @warn  err
                            else
                                throw(DomainError(err))
                            end
                            break
                        else
                            solved=true
                        end
                    end
                    if doprint(control, 'e')
                        @printf("[e]volution:  Δu=%.3e => retry: Δ%s=%.3e\n", Δu, λstr, Δλ)
                    end
                end
            end # while !solved

            if solved 
                istep = istep + 1
                if doprint(control, 'e')
                    @printf(
                        "[e]volution: step=%d %s=%.3e Δ%s=%.3e Δu=%.3e\n",
                        istep,
                        λstr,
                        λ,
                        λstr,
                        Δλ,
                        Δu
                    )
                end
                if control.log
                    push!(allhistory, system.history)
                    push!(allhistory.updates, Δu)
                    push!(allhistory.times, λ)
                end
                if control.store_all
                    append!(tsol, λ, solution)
                end
                control.post(solution, oldsolution, λ, Δλ)
                oldsolution .= solution
                if λ < λend
                    Δλ = min(
                        Δλ_max(control, transient),
                        Δλ * Δλ_grow(control, transient),
                        Δλ * control.Δu_opt / (Δu + 1.0e-14),
                        λend - λ,
                    )
                end
            else 
                break
            end # if solved            
        end # while λ<λ_end

        if !control.store_all # store last solutionobtained
            append!(tsol, λ0, solution)
        end
        control.sample(solution,  λ0)
        if solved
            if !(λ≈lambdas[i+1])
                @warn "λ=$(λ), lambdas[i+1]=$(lambdas[i+1])"
            end
        else
            break
        end
    end # for i = 1:(length(lambdas)-1)

    if doprint(control, 'e')
        println("[e]volution:  $(round(t0+t1,sigdigits=3)) seconds")
    end

    system.history = allhistory
    return tsol
end

"""
    evaluate_residual_and_jacobian(system,u;
                                   t=0.0, tstep=Inf,embed=0.0)

Evaluate residual and jacobian at solution value u.
Returns a solution vector containing the residual, and an ExendableSparseMatrix
containing the linearization at u.

"""
function evaluate_residual_and_jacobian(sys, u; t = 0.0, tstep = Inf, embed = 0.0)
    _complete!(sys; create_newtonvectors = true)

    eval_and_assemble(sys, u, u, sys.residual, t, tstep, embed, zeros(0))
    copy(sys.residual), copy(flush!(sys.matrix))
end

module NoModule end

#####################################################################
"""
    solve(system; kwargs...)
    
Solution method for [`VoronoiFVM.System`](@ref).  
For using ODE solvers from [DifferentialEquations.jl](https://github.com/SciML/DifferentialEquations.jl), see
the [VoronoiFVMDiffEq.jl](https://github.com/j-fu/VoronoiFVMDiffEq.jl) package.
    
Keyword arguments:
- General for all solvers 
   - `inival` (default: 0) : Array created via [`unknowns`](@ref) or  number giving the initial value.
   -  All elements of [`SolverControl`](@ref) can be used as kwargs.
   - `control` (default: nothing): Pass instance of [`SolverControl`](@ref)
   - `params`: Parameters (Parameter handling is experimental and may change)
    
- __Stationary solver__:
  Invoked if neither `times` nor `embed`, nor `tstep` are given as keyword argument.
  - `time` (default: `0.0`): Set time value.
  Returns a [`DenseSolutionArray`](@ref) or [`SparseSolutionArray`](@ref)

- __Embedding (homotopy) solver__: Invoked if `embed` kwarg is given.
  Use homotopy embedding + damped Newton's method  to 
  solve stationary problem or to solve series of parameter dependent problems.
  Parameter step control is performed according to solver control data.  kwargs and default values are:
  - `embed` (default: `nothing` ): vector of parameter values to be reached exactly
  In addition,  all kwargs of the implicit Euler solver (besides `times`) are handled.  
  Returns a transient solution object `sol` containing the stored solution(s),  see [`TransientSolution`](@ref).
  
- __Implicit Euler transient solver__: Invoked if `times` kwarg is given.
  Use implicit Euler method  + damped   Newton's method  to 
  solve time dependent problem. Time step control is performed
  according to solver control data.  kwargs and default values are:
  - `times` (default: `nothing` ): vector of time values to be reached exactly
  - `pre` (default: `(sol,t)->nothing` ):  invoked before each time step
  - `post`  (default:  `(sol,oldsol, t, Δt)->nothing` ):  invoked after each time step
  - `sample` (default:  `(sol,t)->nothing` ): invoked after timestep for all times in `times[2:end]`.
  - `delta` (default:  `(system, u,v,t, Δt)->norm(sys,u-v,Inf)` ):  Value  used to control the time step size `Δu`
  If `control.handle_error` is true, if step solution  throws an error,
  stepsize  is lowered, and  step solution is called again with a smaller time value.
  If `control.Δt<control.Δt_min`, solution is aborted with error.
  Returns a transient solution object `sol` containing the stored solution,  see [`TransientSolution`](@ref).
  
- __Implicit Euler timestep solver__.  Invoked if `tstep` kwarg is given. Solve one time step of the implicit Euler method.
  - `time` (default: `0`): Set time value. 
  - `tstep`: time step
  Returns a [`DenseSolutionArray`](@ref) or [`SparseSolutionArray`](@ref)
"""
function CommonSolve.solve(
    sys::VoronoiFVM.AbstractSystem;
    inival = 0,
    params = zeros(0),
    control = VoronoiFVM.SolverControl(),
    time = 0.0,
    tstep = Inf,
    kwargs...,
)
    fix_deprecations!(control)

    if isa(inival, Number)
        inival = unknowns(sys; inival = inival)
    elseif !VoronoiFVM.isunknownsof(inival, sys)
        @error "wrong type of inival: $(typeof(inival))"
    end

    for pair in kwargs
        if first(pair) != :times &&
           first(pair) != :embed &&
           hasfield(SolverControl, first(pair))
            setproperty!(control, first(pair), last(pair))
        end
    end

    sys.linear_cache = nothing

    if haskey(kwargs, :times) && !isnothing(kwargs[:times])
        solve(
            inival,
            sys,
            kwargs[:times];
            control,
            transient = true,
            params,
            time = kwargs[:times][1],
            called_from_API = true,
        )
    elseif haskey(kwargs, :embed) && !isnothing(kwargs[:embed])
        solve(
            inival,
            sys,
            kwargs[:embed];
            called_from_API = true,
            transient = false,
            control,
            params,
            time,
        )
    else
        solve(inival, sys; called_from_API = true, control, params, time, tstep)
    end
end
